# [S-measure](@id Smeasure_overview)

```@docs
s_measure
```
## Example: random data vs Henon maps

Here, we'll compute the S-measure between random time series (uniform noise), between
time series of a dynamical system (coupled Henon maps).

We start by generating the time series.

```@example smeasure_random_henon
using CausalityTools, DynamicalSystems, Plots, Distributions
npts, Ttr = 10000, 5000
x, y = columns(trajectory(henon2(c_xy = 0.87), npts - 1, Ttr = Ttr))
xr, yr = rand(Uniform(-1, 1), npts), rand(Uniform(-1, 1), npts)
[x y xr yr]
```

Let's plot the time series.

```@example smeasure_random_henon
p_det = plot(xlabel = "", ylabel = "Value", title = "Coupled Henon maps")
plot!(x[1:100], label = "x", marker = stroke(:black), c = :black)
plot!(y[1:100], label = "y", marker = stroke(:red), c = :red)
p_rand = plot(xlabel = "Time step", ylabel = "Value", title = "Random time series")
plot!(xr[1:100], label = "xr", c = :blue)
plot!(yr[1:100], label = "yr", c = :purple)

plot(p_det, p_rand, layout = grid(2, 1), size = (382*2, 400), legend = :bottomright, 
    tickfont = font(13), guidefont = font(13), legendfont = font(13))
savefig("henon_random_timeseries.svg"); nothing # hide
```

![](henon_random_timeseries.svg)

Now we compute the S-measure between the random time series, both from `x` to `y` and from `y` to `x`.
We'll also do the same for the Henon maps. 

The test parameters are embedding dimensions (`dx` for the source and `dy` for the target), the embedding lags (`τx` for the source and `τy` for the target), and the number of nearest neighbors `K`. We'll compute the test with fixed embedding parameters, but a varying number of nearest neighbors (`ks = 2:10`).

For the sake of demonstration, we'll use the same embedding parameters both for the source and for the target, for all analyses. For a real use case, embedding parameters should be chosen more carefully, and will, in general, be different for the source and for the target.

```@example smeasure_random_henon
dx, dy = 5, 5
τx, τy = 1, 1

# Compute the s-measures for different values of k
ks = 2:10
ss_r_xy = [s_measure(xr, yr, dx = dx, τx = τx, dy = dy, τy = τy, K = k) for k in ks]
ss_r_yx = [s_measure(yr, xr, dx = dx, τx = τx, dy = dy, τy = τy, K = k) for k in ks]
ss_henon_xy = [s_measure(x, y, dx = dx, τx = τx, dy = dy, τy = τy, K = k) for k in ks]
ss_henon_yx = [s_measure(y, x, dx = dx, τx = τx, dy = dy, τy = τy, K = k) for k in ks]

plot(xlabel = "# nearest neighbors (k)", ylabel = "S", ylims = (-0.05, 1.05))
plot!(ks, ss_r_xy,  label = "random uncoupled system (x -> y)", marker = stroke(2), c = :black)
plot!(ks, ss_r_yx,  label = "random uncoupled system (y -> x)", marker = stroke(2), c = :red)
plot!(ks, ss_henon_xy, marker = stroke(2), label = "henon unidir (x -> y)")
plot!(ks, ss_henon_yx, marker = stroke(2), label = "henon unidir (y -> x)")
savefig("smeasure_random_henon.svg"); nothing # hide
```

![](smeasure_random_henon.svg)

For uncoupled time series, we expect the value of $S$ to be close to zero. For strongly coupled time series, the value of $S$ should be nonzero and approaching 1. This is exactly what we get: for time random time series, the value of $S$ is close to zero and for the Henon maps, it's clearly non-zero.

Note that the actual dynamical coupling in the Henon system is unidirectional from `x` to `y`. The results (positive $S$ in both directions), however, indicate that the coupling is bidirectional, with the coupling being stronger in one direction. This disagreement between results and ground truth highlights the importance of employing *causal hypothesis testing*. For this, we could use [TimeseriesSurrogates.jl](https://github.com/JuliaDynamics/TimeseriesSurrogates.jl) to 
generate surrogate time series. An in-depth example on how to use surrogate testing can be found in the [transfer entropy](@ref transferentropy) example.
